import 'package:built_collection/built_collection.dart';
import 'package:built_redux/built_redux.dart';
import 'package:built_value/built_value.dart';
import 'package:meta/meta.dart';

import 'reducer_decorators.dart';

part 'list_reducers.g.dart';

@visibleForTesting
void addItemReducer<I>(
  BuiltList<I> state,
  Action<I> action,
  ListBuilder<I> builder,
) {
  assert(!state.contains(action.payload));
  builder.add(action.payload);
}

@visibleForTesting
void clearListReducer<I>(
  BuiltList<I> state,
  Action<Null> action,
  ListBuilder<I> builder,
) {
  builder.clear();
}

ListReducerBuilder<S, B, I>
    makeListReducerBuilder<S extends Built<S, B>, B extends Builder<S, B>, I>(
  BuiltList<I> Function(S state) stateMapper,
  ListBuilder<I> Function(B builder) builderMapper, {
  ActionName<I> addAction,
  ActionName<I> removeAction,
  ActionName<ReplaceItemPayload<I>> replaceAction,
  ActionName<SetIndexPayload<I>> setIndexAction,
  ActionName<MoveItemPayload<I>> moveItemAction,
  ActionName<Iterable<I>> setItemsAction,
  ActionName<Null> clearListAction,
}) {
  assert(builderMapper != null);
  assert(stateMapper != null);
  assert(
    [
      addAction,
      removeAction,
      replaceAction,
      setIndexAction,
      moveItemAction,
      setItemsAction,
      clearListAction,
    ].any((element) => element != null),
    'No action names supplied.',
  );
  final builder = ListReducerBuilder(stateMapper, builderMapper);

  if (addAction != null) {
    builder.add(
      addAction,
      ensureNotNullPayload(addItemReducer),
    );
  }

  if (removeAction != null) {
    builder.add(
      removeAction,
      ensureNotNullPayload(removeItemReducer),
    );
  }

  if (replaceAction != null) {
    builder.add<ReplaceItemPayload<I>>(
      replaceAction,
      ensureNotNullPayload(replaceItemReducer),
    );
  }

  if (setIndexAction != null) {
    builder.add(
      setIndexAction,
      ensureNotNullPayload(setIndexReducer),
    );
  }

  if (moveItemAction != null) {
    builder.add(
      moveItemAction,
      ensureNotNullPayload(moveItemReducer),
    );
  }

  if (setItemsAction != null) {
    builder.add(
      setItemsAction,
      ensureNotNullNorEmptyPayload(setItemsReducer),
    );
  }

  if (clearListAction != null) {
    builder.add(
      clearListAction,
      clearListReducer,
    );
  }

  return builder;
}

@visibleForTesting
void moveItemReducer<I>(
  BuiltList<I> state,
  Action<MoveItemPayload<I>> action,
  ListBuilder<I> builder,
) {
  final itemIndex = state.indexOf(action.payload.item);
  _setIndex(
    itemIndex + action.payload.places,
    action.payload.item,
    state,
    builder,
  );
}

@visibleForTesting
void removeItemReducer<T>(
  BuiltList<T> state,
  Action<T> action,
  ListBuilder<T> builder,
) {
  assert(state.contains(action.payload));
  builder.remove(action.payload);
}

@visibleForTesting
void replaceItemReducer<I>(
  BuiltList<I> state,
  Action<ReplaceItemPayload<I>> action,
  ListBuilder<I> builder,
) {
  assert(state.contains(action.payload.item));
  builder.map((e) => e == action.payload.item ? action.payload.replacement : e);
}

@visibleForTesting
void setIndexReducer<I>(
  BuiltList<I> state,
  Action<SetIndexPayload<I>> action,
  ListBuilder<I> builder,
) {
  _setIndex(
    action.payload.index,
    action.payload.item,
    state,
    builder,
  );
}

@visibleForTesting
void setItemsReducer<I>(
  BuiltList<I> state,
  Action<Iterable<I>> action,
  ListBuilder<I> builder,
) {
  builder.replace(action.payload);
}

void _setIndex<I>(
  int index,
  I item,
  BuiltList<I> state,
  ListBuilder<I> builder,
) {
  assert(index >= 0, 'The index must be greather than zero.');
  assert(index < state.length, 'The index must be less than the length.');
  assert(state.contains(item), 'The item is not present in the list.');
  if (index != state.indexOf(item)) {
    builder.remove(item);
    builder.insert(index, item);
  }
}

abstract class MoveItemPayload<TItem>
    implements Built<MoveItemPayload<TItem>, MoveItemPayloadBuilder<TItem>> {
  factory MoveItemPayload(TItem item, int places) {
    assert(item != null);
    assert(places != null);
    return MoveItemPayload.builder((b) => b
      ..item = item
      ..places = places);
  }
  factory MoveItemPayload.builder(
          [void Function(MoveItemPayloadBuilder<TItem>) updates]) =
      _$MoveItemPayload<TItem>;

  MoveItemPayload._();
  TItem get item;

  int get places;
}

abstract class ReplaceItemPayload<TItem>
    implements
        Built<ReplaceItemPayload<TItem>, ReplaceItemPayloadBuilder<TItem>> {
  factory ReplaceItemPayload(TItem item, TItem replacement) {
    return ReplaceItemPayload.builder((b) => b
      ..item = item
      ..replacement = replacement);
  }
  factory ReplaceItemPayload.builder(
          [void Function(ReplaceItemPayloadBuilder<TItem>) updates]) =
      _$ReplaceItemPayload<TItem>;

  ReplaceItemPayload._();
  TItem get item;

  TItem get replacement;
}

abstract class SetIndexPayload<TItem>
    implements Built<SetIndexPayload<TItem>, SetIndexPayloadBuilder<TItem>> {
  factory SetIndexPayload(int index, TItem item) {
    assert(item != null);
    assert(index != null);
    return SetIndexPayload.builder((b) => b
      ..index = index
      ..item = item);
  }
  factory SetIndexPayload.builder(
          [void Function(SetIndexPayloadBuilder<TItem>) updates]) =
      _$SetIndexPayload<TItem>;

  SetIndexPayload._();
  int get index;

  TItem get item;
}
