import 'package:built_collection/built_collection.dart';
import 'package:built_redux/built_redux.dart';
import 'package:built_redux_extras/built_redux_extras.dart';
import 'package:built_value/built_value.dart';
import 'package:test/test.dart';

part 'list_reducer_test.g.dart';

void main() {
  group('List reducers', () {
    final storeReducer = (ReducerBuilder<_StoreState, _StoreStateBuilder>()
          ..combineList(
              makeListReducerBuilder<_StoreState, _StoreStateBuilder, int>(
            (state) => state.numbers,
            (builder) => builder.numbers,
            addAction: _ListActionsNames.add,
            clearListAction: _ListActionsNames.clear,
            moveItemAction: _ListActionsNames.moveItem,
            removeAction: _ListActionsNames.remove,
            replaceAction: _ListActionsNames.replaceItem,
            setIndexAction: _ListActionsNames.setIndex,
            setItemsAction: _ListActionsNames.setItems,
          )))
        .build();
    final actions = _ListActions();
    Store<_StoreState, _StoreStateBuilder, _ListActions> store;

    setUp(() {
      store = Store(storeReducer, _StoreState([1, 3, 2, 5, 7]), actions);
    });

    tearDown(() {
      store.dispose();
      store = null;
    });

    test('Add item action', () {
      actions.add(4);
      actions.add(6);
      expect(store.state.numbers, containsAll([4, 6]));
    });

    test('Try to add an existing item', () {
      expect(
        () => actions.add(5),
        throwsA(isA<AssertionError>()),
      );
    });

    test('Item is removed', () {
      actions.remove(7);
      expect(store.state.numbers, isNot(contains(7)));
    });

    test('Try to remove nonexistent item', () {
      expect(
        () => actions.remove(4),
        throwsA(isA<AssertionError>()),
      );
    });

    test('Replace item action', () {
      actions.replaceItem(ReplaceItemPayload(7, 9));
      expect(store.state.numbers, isNot(contains(7)));
      expect(store.state.numbers, contains(9));
    });

    test('Try to replace a nonexistent item', () {
      expect(
        () => actions.replaceItem(ReplaceItemPayload(4, 9)),
        throwsA(isA<AssertionError>()),
      );
    });

    test('Set item index action', () {
      actions.setIndex(SetIndexPayload(1, 2));
      expect(store.state.numbers, containsAllInOrder([1, 2, 3]));
    });

    test('Set item index for a nonexistent item', () {
      expect(
        () => actions.setIndex(SetIndexPayload(1, 4)),
        throwsA(isA<AssertionError>()),
      );
    });

    test('Set index to less than zero', () {
      expect(() {
        actions.setIndex(SetIndexPayload(-2, 2));
      }, throwsA(isA<AssertionError>()));
    });

    test('Set index to greater than length', () {
      expect(() {
        actions.setIndex(SetIndexPayload(10, 2));
      }, throwsA(isA<AssertionError>()));
    });

    test('Move item action down', () {
      actions.moveItem(MoveItemPayload(1, 1));
      expect(store.state.numbers, containsAllInOrder([3, 1, 2]));
    });

    test('Move item action up', () {
      actions.moveItem(MoveItemPayload(2, -1));
      expect(store.state.numbers, containsAllInOrder([1, 2, 3]));
    });

    test('Set items action', () {
      actions.setItems([1, 2, 3, 4]);
      expect(store.state.numbers, containsAllInOrder([1, 2, 3, 4]));
    });

    test('Clear list action', () {
      actions.clear();
      expect(store.state.numbers, isEmpty);
    });
  });
}

abstract class _StoreState implements Built<_StoreState, _StoreStateBuilder> {
  BuiltList<int> get numbers;

  _StoreState._();
  factory _StoreState.builder([void Function(_StoreStateBuilder) updates]) =>
      _$StoreState(updates);

  factory _StoreState([Iterable<int> numbers = const []]) =>
      _StoreState.builder((b) => b.numbers.replace(numbers));
}

abstract class _ListActions extends ReduxActions {
  _ListActions._();
  factory _ListActions() => _$_ListActions();

  ActionDispatcher<int> get add;
  ActionDispatcher<int> get remove;
  ActionDispatcher<ReplaceItemPayload<int>> get replaceItem;
  ActionDispatcher<SetIndexPayload<int>> get setIndex;
  ActionDispatcher<MoveItemPayload<int>> get moveItem;
  ActionDispatcher<Iterable<int>> get setItems;
  ActionDispatcher<Null> get clear;
}
